"""neodym.client

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import asyncore
import logging
import socket

from neodym.connection import Connection
from neodym.message import Message


class Client(asyncore.dispatcher):
    """neodym.client.Client

    The client connects to a server and tries to establish a connection by
    performing a handshaking procedure. If the handshake completes
    successfully, the client can send messages to the server.
    """
    __hash__ = ""

    def __init__(self, host, port):
        """
        :param host:    hostname or ip of server as string
        :param port:    TCP port as integer
        """
        self.host = host
        self.port = port
        asyncore.dispatcher.__init__(self)

        self.logger = logging.getLogger('Client-%s' % id(self))
        self.logger.debug('Initializing: %s' % self)

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

        self.__connection = None

    def handle_connect(self):
        """asyncore.handle_connect
        """
        self.logger.info('Connected to server, establishing connection...')
        self.__connection = Connection(self.socket, self)

    def handle_close(self):
        """asyncore.handle_close
        """
        self.logger.info('Closing.')
        Connection.__all__ = set()

    def client_connect(self):
        """neodym.client.Client.client_connect

        Connect to server.
        """
        self.connect((self.host, self.port))
        pass
        self.update()
        pass

    def update(self):
        """neodym.client.Client.update

        Poll asyncore, pass and return.
        """
        asyncore.poll2()
        pass
        return

    def connection(self):
        """neodym.client.Client.connection

        Obtain an established connection from the client by performing the
        handshaking procedure with the server.

        :returns neodym.connection.Connection:  the connection instance
        """
        handshake = Message('handshake', [self.__hash__])
        self.__connection.put(handshake)

        while self.__connection.recv_queue.empty():
            self.update()

        message = self.__connection.recv_queue.get()
        if not message:
            self.logger.info('Timeout reached!')
            self.close()
        elif message.unique_identifier == 'handshake':
            self.logger.info('Connection established.')
            return self.__connection



"""neodym.exception

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

class NeodymException(Exception):
    pass

class NotYetInitialized(NeodymException):
    """Neodym is not yet initialized"""
    def __init__(self):
        NeodymException.__init__(self, self.__doc__)

class AlreadyInitialized(NeodymException):
    """Neodym is already initialized!"""
    def __init__(self):
        NeodymException.__init__(self, self.__doc__)

### Messages ###
class NeodymMessageException(NeodymException):
    pass

class UnregisteredMessage(NeodymMessageException):
    """All transport messages need to be registered before initializing Neodym."""
    def __init__(self):
        NeodymMessageException.__init__(self, self.__doc__)

class MalformedMessage(NeodymMessageException):
    """The message is malformed or invalid"""
    def __init__(self):
        NeodymMessageException.__init__(self, self.__doc__)

class UnknownAttribute(NeodymMessageException):
    """The attribute you are requesting is not an attribute of this message"""
    def __init__(self):
        NeodymMessageException.__init__(self, self.__doc__)



"""neodym.message.Message

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import logging
import UserDict
import hashlib
import json

from neodym.exceptions import (AlreadyInitialized, MalformedMessage,
                               UnregisteredMessage, UnknownAttribute)


class RegisteredMessageDict(UserDict.IterableUserDict):
    """neodym.message.RegisteredMessageDict

    This is the message-mapper holding all registered messages of this neodym instance.

    The message-mapper behaves like a regular dictionary, except that it has an extra
    method for generating an md5 hash over all registered messages for the handshaking
    procedure.

    The message-mapper can be initialized in three different ways:

        - msg_map = neodym.message.RegisteredMessageDict()
        - msg_map = neodym.message.RegisteredMessageDict({'foo': 'bar'})
        - msg_map = neodym.message.RegisteredMessageDict(foo=bar)

    There should always exist only one message-mapper!
    """
    # todo: refactor towards singleton
    __hash__ = ""

    def __init__(self, *args, **kwargs):
        self.data = {}

        self.logger = logging.getLogger('RegisteredMessageDict')
        self.logger.info('Initializing: %s' % self)

        UserDict.IterableUserDict.__init__(self, *args, **kwargs)

    def __setitem__(self, key, value):
        """neodym.message.RegisteredMessageDict.__setitem__

        Items can only be set while neodym is not yet initialized.

        :raises: AlreadyInitialized:    if neodym is already initialized
        """
        if self.__hash__:
            raise AlreadyInitialized

        if not isinstance(key, str) or not isinstance(value, list):
            raise ValueError('Key must be a string and value must be a list!')
        self.data[key] = value

    def __delitem__(self, key):
        """neodym.message.RegisteredMessageDict.__delitem__

        Items can only be deleted while neodym has not yet been initialized.

        :raises AlreadyInitialized:     if neodym has already been initialized
        """
        if self.__hash__:
            raise AlreadyInitialized

    def hash(self):
        """neodym.message.RegisteredMessageDict.hash

        Return the md5 hash for all currently registered messages.

        :returns str:   md5 hex-digest
        """
        msg_map = [(k, self.data[k]) for k in sorted(self.data)]
        return hashlib.md5(str(msg_map)).hexdigest()


class Message(object):
    """neodym.message.Message

    Messages are the core elements of neodym network communications. They are
    the atom element that is sent over the wire.

    All messages need to be registered before initializing neodym. Only clients
    with matching message-mapper hashes are allowed to connect with the server.
    This measure insures that both parties are able to interpret all possible
    messages equally.

    Only messages that have been registered with neodym will be able to be
    initialized. Initializing a message that hasn't been registered with neodym
    will result in an neodym.exceptions.MalformedMessage.

    A message consists of a unique_identifier and a list of positional
    attributes. The length of the positional attributes must be equal to the
    length of the positional attribute-names in the message-mapper, and
    additionally in the identical order!
    """
    __map__ = dict()    # dictionary of all registered messages

    def __init__(self, unique_identifier, attrs):
        """
        :param unique_identifier:   unique-identifier of message as string
        :param attrs:               list of positional attributes of message
        """
        self.unique_identifier = unique_identifier
        self.attrs = attrs

        self.logger = logging.getLogger('Message-%s' % id(self))
        self.logger.debug('Initializing: %s' % self)

        if not (
            isinstance(self.unique_identifier, str) and
            isinstance(self.attrs, list)
        ):
            raise MalformedMessage

        if not (
            self.unique_identifier in self.__map__ and
            len(self.attrs) == len(self.__map__[self.unique_identifier])
        ):
            raise UnregisteredMessage

        if not self.unique_identifier in self.__map__:
            raise UnregisteredMessage

    def __repr__(self):
        """neodym.message.Message.__repr__

        Print a nice representation of this message.
        """
        return "<Neodym.Message(%s)>" % self.pack()[:-2]

    def get_attr(self, attr):
        """neodym.message.Message.get_attr

        Get a specific message-attribute. Will raise an exception if the
        attribute doesn't exist.

        :returns attr:

        :raises neodym.exceptions.UnknownAttribute:     if attr is unknown
        """
        if not attr in self.__map__[self.unique_identifier]:
            raise UnknownAttribute

        index = self.__map__[self.unique_identifier].index(attr)
        return self.attrs[index]

    def set_attr(self, attr, value):
        """neodym.message.Message.set_attr

        Set a specific message-attribute to a given value. Will raise an
        exception if the attribute doesn't exist.

        :raises neodym.exceptions.UnknownAttribute:     if attr is unknown
        """
        if not attr in self.__map__[self.unique_identifier]:
            raise UnknownAttribute

        index = self.__map__[self.unique_identifier].index(attr)
        self.attrs[index] = value

    def get_attrs(self):
        """neodym.message.Message.get_attrs

        Return a list of tuples of all attribute-names and their values, e.g.
        [('my_attr_name', 'my_attr_value), ... ]

        Hint:   dict(neodym.message.Message.get_attrs) will return a dict of
                all message-attributes mapped to their attribute-names.
        """
        return zip(self.__map__[self.unique_identifier], self.attrs)

    def set_attrs(self, attrs):
        """neodym.message.Message.set_attrs

        Bulk-set many message-attributes to given values. Will raise an
        exception if any of the attribute-names doesn't exist.

        :param attrs:   dictionary containing attribute-names and -values
        """
        if not isinstance(attrs, dict):
            raise TypeError('Attributes must be passed as dictionary!')

        for attr, value in attrs.items():
            self.set_attr(attr, value)

    def pack(self):
        """neodym.message.Message.pack

        Return JSON representation of this message with trailing
        carriage-return and newline.
        """
        return json.dumps({self.unique_identifier: self.attrs}) + '\r\n'

    @classmethod
    def unpack(cls, line):
        """neodym.message.Message.unpack

        Class-method for unpacking a given string into a
        neodym.message.Message. Will return None if the message couldn't be
        unpacked.

        :param line:    arbitrary string

        :returns neodym.message.Message:    if message could be unpacked
        """
        try:
            message = json.loads(line.strip())
            unique_identifier, attrs = message.items()[0]
            return Message(str(unique_identifier), attrs)
        except:
            logging.debug('Error while unpacking line: %s' % line)




"""neodym.server

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import asyncore
import logging
import socket
import time

from neodym.connection import Connection
from neodym.exceptions import NotYetInitialized
from neodym.handler import Handler


class Server(asyncore.dispatcher):
    """neodym.server.Server

    The server listens on an address and waits for connection-request from
    clients. If a connection-request is received, a new instance of
    neodym.connection.Connection is initialized and the client socket is passed
    along. The connection can be obtained by calling the connection method.
    """
    __hash__ = ""
    __max__ = 3

    def __init__(self, address):
        """
        :param address:     address-tuple in format of ('127.0.0.1', 0)
        """
        asyncore.dispatcher.__init__(self)

        self.logger = logging.getLogger('Server-%s' % id(self))
        self.logger.debug('Initializing: %s' % self)

        if not self.__hash__:
            raise NotYetInitialized

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.bind(address)
        self.address = self.socket.getsockname()

        self.set_reuse_addr()

    def __del__(self):
        """neodym.server.Server.__del__

        Leave a note in the neodym logs if the server vanishes.
        """
        self.logger.debug('Vanishing!')

    def server_activate(self):
        """neodym.server.Server.server_activate

        Active the server, e.g. start listening.
        """
        self.logger.info('Binding on %s (max: %s)' % (self.address,
                                                       self.__max__))
        self.listen(self.__max__)

    def handle_accept(self):
        """asyncore.dispatcher.handle_accept
        """
        self.logger.info('Connection request incoming...')
        client_info = self.accept()
        if client_info:
            socket, port = client_info
            Connection(socket, self)

    def handle_close(self):
        """asyncore.dispatcher.handle_close"""
        self.logger.info('Closing.')
        self.close()

    def update(self):
        """neodym.server.Server.update

        Call asyncore.poll2 and then iterate over all connections and handle
        any given input.
        """
        asyncore.poll2()
        pass

        connections = [c for c in Connection.__all__]
        if connections:
            for c in connections:
                if not c.recv_queue.empty():
                    message = c.recv_queue.get()
                    self.logger.debug('Received message: %s' % message)

                    # perform handshake operation on new connections
                    if c.is_connected is False:
                        if message.unique_identifier == 'handshake':
                            self.logger.debug('Handling handshake request...')
                            if message.get_attr('msg_map_hash') == self.__hash__:
                                self.logger.info('Handshake: OK -> Connection-%s' % id(c))
                                c.is_connected = True
                                c.put(message)
                                c.handle_write()
                            else:
                                self.logger.info('Hash mismatch!')

                    # handle the message for all connected clients
                    elif c.is_connected is True:
                        handlers = Handler.get_handlers(message)
                        callbacks = [h(message, c) for h in handlers]
                        self.logger.debug('Handling message through: %s' % (
                            str(callbacks)
                        ))
                    else:
                        self.logger.debug('Internal server error!')
        else:
            time.sleep(0.1)

    def serve_forever(self):
        """neodym.server.Server.serve_forever

        Update the server in a continuous loop.
        """
        while True:
            self.update()



"""Neodym

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import client
import server
import connection
import message
import handler
import exceptions
import logging

# message-mapper for all registered messages
msg_map = message.RegisteredMessageDict(
    handshake=['msg_map_hash']
)

# some shortcuts for our logging wrapper
log_level = logging.INFO
log_format = '%(asctime)16s Neodym %(levelname)10s %(name)32s - %(message)s'


def baseConfig():
    """neodym.baseConfig

    Helper for obtaining a usable logging.baseConfig

    :returns dict:      dictionary with logging configuration
    """
    return {
        'level': log_level,
        'format': log_format,
    }


def init():
    """neodym.init

    Neodym has an initialization and handshaking procedure ensuring both sides
    (e.g. server and all clients) have registered the exactly identical
    message-definitions - see: help(neodym.register) for more information.

    Calling neodym.init() will freeze the registration of new
    message-definitions, generate an md5 hash from neodym.msg_map and pass this hash
    down to all neodym objects.

    Neither client nor server can be initialized before initializing neodym!
    (see: help(neodym.Client) and help(neodym.Server) for more information)
    """
    msg_map_hash = msg_map.hash()
    msg_map.__hash__ = msg_map_hash
    server.Server.__hash__ = msg_map_hash
    client.Client.__hash__ = msg_map_hash
    connection.Connection.__hash__ = msg_map_hash
    message.Message.__map__ = msg_map
    handler.Handler.__map__ = msg_map


def register(unique_identifier, attrs):
    """neodym.register

    Register a new message-definition in the message-mapper. The message-mapper
    is a helper that holds all registered message-definitions.

    A message consists of a unique-identifier and a list of positional
    arguments - see: help(neodym.Message) for more information.

    :param unique_identifier:   unique-identifier of message as string
    :param attrs:               all positional arguments of message as list
    """
    msg_map[unique_identifier] = attrs


# some import shortcuts
from client import Client
from server import Server
from message import Message
from handler import Handler



"""neodym.handler

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org>"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import logging

from neodym.exceptions import NotYetInitialized, UnregisteredMessage
from neodym.message import Message


class Handler(object):
    """neodym.handler.Handler

    Handlers are helpers for organizing and managing callbacks. This class acts
    as a base-class for all custom handlers you may want to define.

    Handlers have a unique-identifier that must match a unique-identifier of a
    registered message. This identifier indicates what message this handler is
    registered to. Several handlers may be attached to a message, they will be
    called in the order they have been registered when the corresponding
    message arrives.
    """
    __all__ = dict()
    __map__ = dict()

    def __init__(self, unique_identifier):
        """
        :param unique_identifier:   unique-identifier of message this handler
                                    is registered to handle.
        """
        self.unique_identifier = unique_identifier

        self.logger = logging.getLogger('Handler-%s' % id(self))
        self.logger.debug('Initializing: %s' % self)

        if not self.__map__:
            raise NotYetInitialized

        if not self.unique_identifier in self.__map__:
            raise UnregisteredMessage

        if not self.unique_identifier in self.__all__:
            self.__all__[self.unique_identifier] = [self]
        else:
            self.__all__[self.unique_identifier].append(self)

    def __repr__(self):
        """neodym.handler.Handler.__repr__

        Print a nice representation for this handler
        """
        return "<Handler(%s)>" % self.unique_identifier

    def __call__(self, message, connection):
        """neodym.handler.Handler.__call__

        Call this handler.

        :param message:     instance of neodym.message.Message
        :param connection:  instance of neodym.connection.Connection
        """
        self.logger.debug('Received call-back.')
        if not isinstance(message, Message):
            raise TypeError('must be of type neodym.message.Message')

        self.handle(message, connection)
        return self

    def handle(self, message, connection):
        """Overwrite this method in custom handler.
        """
        pass

    def drop(self):
        """neodym.handler.Handler.drop

        Drop this handler so it wont be called on messages any more.
        """
        del self.__all__[self.unique_identifier]

    @classmethod
    def get_handlers(cls, message):
        """neodym.handler.Handler.get_handlers

        Get the list of handlers for a given message.

        :param message:     instance of neodym.message.Message
        """
        if isinstance(message, Message):
            if message.unique_identifier in cls.__map__:
                return cls.__all__[message.unique_identifier]



"""neodym.connection

Neodym - A thin message-bus wrapper around asyncore.
Copyright (C) 2013  Brian Wiborg <baccenfutter@c-base.org>

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License as published by the Free
Software Foundation, either version 3 of the License, or (at your option) any
later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Affero General Public License for more
details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""
__author__ = "Brian Wiborg <baccenfutter@c-base.org"
__date__ = "2013/08/31"
__license__ = "GNU/AGPL-3.0"

import asyncore
import logging
import Queue

from neodym.exceptions import MalformedMessage
from neodym.message import Message


class Connection(asyncore.dispatcher):
    """neodym.connection.Connection

    A connection is an asyncore.dispatcher acting as a communication interface
    between client and server. From client-view, the connection is the
    representation of the server. From server-view, each connection represents
    a connected client.
    """
    __all__ = []
    __hash__ = ""
    __chunk__ = 8192
    __timeout__ = 30

    def __init__(self, sock, parent):
        """
        :param sock:    socket.socket
        :param parent:  instance either of neodym.server.Server or
                        neodym.client.Client
        """
        asyncore.dispatcher.__init__(self, sock=sock)
        self.parent = parent

        self.logger = logging.getLogger('Connection-%s' % id(self))
        self.logger.debug('Initializing: %s' % self)

        self.send_queue = Queue.Queue()
        self.recv_queue = Queue.Queue()

        self.is_connected = False
        Connection.__all__.append(self)

    def __del__(self):
        """neodym.connection.Connection.__del__

        Leave a short note in the server logs about the terminating connection.
        """
        self.logger.info('Terminated.')

    def writable(self):
        """asyncore.dispatcher.writable
        """
        return not self.send_queue.empty()

    def handle_read(self):
        """asyncore.dispatcher.handle_read
        """
        self.logger.debug('Data incoming...')
        chunk = self.recv(self.__chunk__)
        if not chunk:
            self.close()

        lines = chunk.replace('\r\n', '\n').split('\n')
        self.logger.debug('Received data: %s' % str(lines))

        for line in lines:
            if line:
                message = Message.unpack(line)
                if message:
                    self.recv_queue.put(message)
                else:
                    self.logger.debug('Can not unpack message: %s' % str(message))

    def handle_write(self):
        """asyncore.dispatcher.handle_write
        """
        message = self.send_queue.get()
        self.logger.debug('Handling write: %s' % message)
        if isinstance(message, Message):
            self.send(message.pack())
        else:
            self.send(str(message) + '\r\n')

    def handle_close(self):
        """asyncore.dispatcher.handle_close
        """
        self.logger.debug('Closing!')
        self.drop()

    def put(self, message):
        """neodym.connection.Connection.put

        Put a message into the send-queue.

        :param message:     instance of neodym.message.Message
        """
        self.logger.debug('Throwing message into send queue: %s' % message)
        if not isinstance(message, Message):
            raise MalformedMessage

        self.send_queue.put(message)

    def get(self):
        """neodym.connection.Connection.get

        Poll the receive-queue for any new messages that may have arrived.

        :returns neodym.message.Message:    if a message is in the queue
        :returns None                       if queue is empty
        """
        self.logger.debug('Polling receive queue for messages')
        if not self.recv_queue.empty():
            return self.recv_queue.get()

    def drop(self):
        """neodym.connection.Connection.drop

        Drop this connection from the list of all connections, e.g. as a
        preparation for closing the connection.
        """
        if self in self.__all__:
            self.__all__.remove(self)



from distutils.core import setup

setup(
    name='NeoDym',
    version='0.2.1',
    author='Brian Wiborg',
    author_email='baccenfutter@c-base.org',
    packages=['neodym'],
    scripts=[],
    url='http://pypi.python.org/pypi/NeoDym/',
    license='LICENSE.txt',
    description='A thin message-bus wrapper around asyncore.',
    long_description=open('README.txt').read(),
    install_requires=[],
)

